@@ -41,8 +41,14 @@
 #include "inet.h"
 #include "supplicant.h"
 
+#define CLEANUP_TIMEOUT   8	/* in seconds */
+#define INACTIVE_TIMEOUT  12	/* in seconds */
+
 struct wifi_data {
-	GSList *list;
+	GSList *current;
+	GSList *pending;
+	guint cleanup_timer;
+	guint inactive_timer;
 	gchar *identifier;
 	gboolean connected;
 };
@@ -61,29 +67,30 @@ static void network_remove(struct connman_element *element)
 
 static int network_enable(struct connman_element *element)
 {
-	char *identifier, *security = NULL, *passphrase = NULL;
+	struct connman_element *device = element->parent;
+	char *name, *security = NULL, *passphrase = NULL;
 	unsigned char *ssid;
 	int ssid_len;
 
 	DBG("element %p name %s", element, element->name);
 
 	if (connman_element_get_static_property(element,
-					"WiFi.Name", &identifier) == FALSE)
+						"Name", &name) == FALSE)
 		return -EIO;
 
 	if (connman_element_get_static_array_property(element,
 				"WiFi.SSID", &ssid, &ssid_len) == FALSE)
 		return -EIO;
 
-	if (element->parent != NULL) {
-		struct wifi_data *data = connman_element_get_data(element->parent);
-
-		if (data->connected == TRUE)
-			return -EBUSY;
+	if (device != NULL) {
+		struct wifi_data *data = connman_element_get_data(device);
 
 		if (data != NULL) {
+			if (data->connected == TRUE)
+				return -EBUSY;
+
 			g_free(data->identifier);
-			data->identifier = g_strdup(identifier);
+			data->identifier = g_strdup(name);
 		}
 	}
 
@@ -93,8 +100,8 @@ static int network_enable(struct connman_element *element)
 	connman_element_get_value(element,
 			CONNMAN_PROPERTY_ID_WIFI_PASSPHRASE, &passphrase);
 
-	DBG("identifier %s security %s passhprase %s",
-					identifier, security, passphrase);
+	DBG("name %s security %s passhprase %s",
+					name, security, passphrase);
 
 	if (__supplicant_connect(element, ssid, ssid_len,
 						security, passphrase) < 0)
@@ -124,41 +131,107 @@ static struct connman_driver network_driver = {
 	.disable	= network_disable,
 };
 
-static struct connman_element *find_element(struct wifi_data *data,
-						const char *identifier)
+static struct connman_element *find_current_element(struct wifi_data *data,
+							const char *identifier)
 {
 	GSList *list;
 
-	for (list = data->list; list; list = list->next) {
+	for (list = data->current; list; list = list->next) {
 		struct connman_element *element = list->data;
 
 		if (connman_element_match_static_property(element,
-					"WiFi.Name", &identifier) == TRUE)
+					"Name", &identifier) == TRUE)
 			return element;
 	}
 
 	return NULL;
 }
 
-static void state_change(struct connman_element *parent,
+static struct connman_element *find_pending_element(struct wifi_data *data,
+							const char *identifier)
+{
+	GSList *list;
+
+	for (list = data->pending; list; list = list->next) {
+		struct connman_element *element = list->data;
+
+		if (connman_element_match_static_property(element,
+					"Name", &identifier) == TRUE)
+			return element;
+	}
+
+	return NULL;
+}
+
+static gboolean inactive_scan(gpointer user_data)
+{
+	struct connman_element *device = user_data;
+	struct wifi_data *data = connman_element_get_data(device);
+
+	DBG("");
+
+	if (data->cleanup_timer > 0) {
+		g_source_remove(data->cleanup_timer);
+		data->cleanup_timer = 0;
+	}
+
+	__supplicant_scan(device);
+
+	data->inactive_timer = 0;
+
+	return FALSE;
+}
+
+static void connect_known_networks(struct connman_element *device)
+{
+	struct wifi_data *data = connman_element_get_data(device);
+	GSList *list;
+
+	DBG("");
+
+	if (data->inactive_timer > 0) {
+		g_source_remove(data->inactive_timer);
+		data->inactive_timer = 0;
+	}
+
+	for (list = data->current; list; list = list->next) {
+		struct connman_element *element = list->data;
+
+		if (element->policy == CONNMAN_ELEMENT_POLICY_AUTO &&
+						element->remember == TRUE &&
+						element->available == TRUE) {
+			if (network_enable(element) == 0)
+				return;
+		}
+	}
+
+	data->inactive_timer = g_timeout_add_seconds(INACTIVE_TIMEOUT,
+							inactive_scan, device);
+}
+
+static void state_change(struct connman_element *device,
 						enum supplicant_state state)
 {
-	struct wifi_data *data = connman_element_get_data(parent);
+	struct wifi_data *data = connman_element_get_data(device);
 	struct connman_element *element;
 
 	DBG("state %d", state);
 
-	if (data->identifier == NULL)
+	if (data == NULL)
 		return;
 
-	element = find_element(data, data->identifier);
+	if (data->identifier == NULL)
+		goto reconnect;
+
+	element = find_current_element(data, data->identifier);
 	if (element == NULL)
-		return;
+		goto reconnect;
 
 	if (state == STATE_COMPLETED) {
 		struct connman_element *dhcp;
 
 		data->connected = TRUE;
+		connman_element_set_enabled(element, TRUE);
 
 		dhcp = connman_element_create(NULL);
 
@@ -166,15 +239,67 @@ static void state_change(struct connman_element *parent,
 		dhcp->index = element->index;
 
 		connman_element_register(dhcp, element);
-	} else if (state == STATE_DISCONNECTED || state == STATE_INACTIVE)
+	} else if (state == STATE_INACTIVE || state == STATE_DISCONNECTED) {
 		data->connected = FALSE;
+		connman_element_set_enabled(element, FALSE);
+
+		connman_element_unregister_children(element);
+	}
+
+reconnect:
+	if (state == STATE_INACTIVE || state == STATE_DISCONNECTED) {
+		data->connected = FALSE;
+		connect_known_networks(device);
+	}
 }
 
-static void scan_result(struct connman_element *parent,
+static gboolean cleanup_pending(gpointer user_data)
+{
+	struct wifi_data *data = user_data;
+	GSList *list;
+
+	DBG("");
+
+	for (list = data->pending; list; list = list->next) {
+		struct connman_element *element = list->data;
+
+		DBG("element %p name %s", element, element->name);
+
+		connman_element_unregister(element);
+		connman_element_unref(element);
+	}
+
+	g_slist_free(data->pending);
+	data->pending = NULL;
+
+	data->cleanup_timer = 0;
+
+	return FALSE;
+}
+
+static void clear_results(struct connman_element *device)
+{
+	struct wifi_data *data = connman_element_get_data(device);
+
+	DBG("pending %d", g_slist_length(data->pending));
+	DBG("current %d", g_slist_length(data->current));
+
+	data->pending = data->current;
+	data->current = NULL;
+
+	if (data->cleanup_timer > 0)
+		return;
+
+	data->cleanup_timer = g_timeout_add_seconds(CLEANUP_TIMEOUT,
+							cleanup_pending, data);
+}
+
+static void scan_result(struct connman_element *device,
 					struct supplicant_network *network)
 {
-	struct wifi_data *data = connman_element_get_data(parent);
+	struct wifi_data *data = connman_element_get_data(device);
 	struct connman_element *element;
+	guint8 strength = network->quality;
 	gchar *temp;
 	int i;
 
@@ -192,31 +317,20 @@ static void scan_result(struct connman_element *parent,
 	temp = g_strdup(network->identifier);
 
 	for (i = 0; i < strlen(temp); i++) {
-		if (temp[i] == ' ' || temp[i] == '.')
+		gchar tmp = g_ascii_tolower(temp[i]);
+
+		if (tmp < 'a' || tmp > 'z')
 			temp[i] = '_';
-		else if (temp[i] == '-' || temp[i] == '+')
-			temp[i] = '_';
-		else if (temp[i] == '!' || temp[i] == '?')
-			temp[i] = '_';
-		else if (temp[i] == '(' || temp[i] == ')')
-			temp[i] = '_';
-		else if (g_ascii_isprint(temp[i]) == FALSE)
-			temp[i] = '_';
-		temp[i] = g_ascii_tolower(temp[i]);
 	}
 
-	element = find_element(data, network->identifier);
+	element = find_pending_element(data, network->identifier);
 	if (element == NULL) {
-		guint8 strength;
-
 		element = connman_element_create(temp);
 
 		element->type = CONNMAN_ELEMENT_TYPE_NETWORK;
-		element->index = parent->index;
+		element->index = device->index;
 
-		data->list = g_slist_append(data->list, element);
-
-		connman_element_add_static_property(element, "WiFi.Name",
+		connman_element_add_static_property(element, "Name",
 				DBUS_TYPE_STRING, &network->identifier);
 
 		connman_element_add_static_array_property(element, "WiFi.SSID",
@@ -237,25 +351,30 @@ static void scan_result(struct connman_element *parent,
 			element->wifi.security = g_strdup(security);
 		}
 
-		strength = network->quality;
-
-		//connman_element_add_static_property(element, "WiFi.Strength",
-		//				DBUS_TYPE_BYTE, &strength);
-
-		//connman_element_add_static_property(element, "WiFi.Noise",
-		//			DBUS_TYPE_INT32, &network->noise);
+		connman_element_add_static_property(element, "Strength",
+						DBUS_TYPE_BYTE, &strength);
 
 		DBG("%s (%s) strength %d", network->identifier,
 					element->wifi.security, strength);
 
-		connman_element_register(element, parent);
+		connman_element_register(element, device);
+	} else {
+		data->pending = g_slist_remove(data->pending, element);
+
+		connman_element_set_static_property(element, "Strength",
+						DBUS_TYPE_BYTE, &strength);
 	}
 
+	data->current = g_slist_append(data->current, element);
+
+	element->available = TRUE;
+
 	g_free(temp);
 }
 
 static struct supplicant_callback wifi_callback = {
 	.state_change	= state_change,
+	.clear_results	= clear_results,
 	.scan_result	= scan_result,
 };
 
@@ -319,19 +438,40 @@ static int wifi_disable(struct connman_element *element)
 
 	DBG("element %p name %s", element, element->name);
 
+	if (data->cleanup_timer > 0) {
+		g_source_remove(data->cleanup_timer);
+		data->cleanup_timer = 0;
+	}
+
+	if (data->inactive_timer > 0) {
+		g_source_remove(data->inactive_timer);
+		data->inactive_timer = 0;
+	}
+
 	__supplicant_disconnect(element);
 
-	for (list = data->list; list; list = list->next) {
+	for (list = data->pending; list; list = list->next) {
 		struct connman_element *network = list->data;
 
 		connman_element_unref(network);
 	}
 
-	g_slist_free(data->list);
-	data->list = NULL;
+	g_slist_free(data->pending);
+	data->pending = NULL;
+
+	for (list = data->current; list; list = list->next) {
+		struct connman_element *network = list->data;
+
+		connman_element_unref(network);
+	}
+
+	g_slist_free(data->current);
+	data->current = NULL;
 
 	connman_element_unregister_children(element);
 
+	__supplicant_stop(element);
+
 	return 0;
 }
 
@@ -354,7 +494,7 @@ static void wifi_newlink(unsigned short type, int index,
 	struct connman_element *device;
 	GSList *list;
 	gboolean exists = FALSE;
-	gchar *name;
+	gchar *name, *devname;
 	struct iwreq iwr;
 	int sk;
 
@@ -363,10 +503,11 @@ static void wifi_newlink(unsigned short type, int index,
 	if (type != ARPHRD_ETHER)
 		return;
 
-	name = inet_index2name(index);
+	name = inet_index2ident(index, "dev_");
+	devname = inet_index2name(index);
 
 	memset(&iwr, 0, sizeof(iwr));
-	strncpy(iwr.ifr_ifrn.ifrn_name, name, IFNAMSIZ);
+	strncpy(iwr.ifr_ifrn.ifrn_name, devname, IFNAMSIZ);
 
 	sk = socket(PF_INET, SOCK_DGRAM, 0);
 
@@ -398,6 +539,7 @@ static void wifi_newlink(unsigned short type, int index,
 
 	device->index = index;
 	device->name = name;
+	device->devname = devname;
 
 	connman_element_register(device, NULL);
 	device_list = g_slist_append(device_list, device);
@@ -507,5 +649,5 @@ static void wifi_exit(void)
 	dbus_connection_unref(connection);
 }
 
-CONNMAN_PLUGIN_DEFINE("wifi", "WiFi interface plugin", VERSION,
+CONNMAN_PLUGIN_DEFINE(wifi, "WiFi interface plugin", VERSION,
 							wifi_init, wifi_exit)
